import re

from cfgformalizer.cisco.statement import Statement
from cfgformalizer.stanza import Stanza


def depth(str):
    return len(str) - len(str.lstrip(" "))


class Cisco:
    @classmethod
    def parse(self, string):
        stanza = []
        prev_line = ""
        prev_line_depth = 0
        context = []
        current_context_depth = 0
        banner_delimiter = None
        linenum = 1
        seqnum = 0
        stanza = Stanza()

        banner_regexp = re.compile(r"(?:(?:banner(?:\smotd)?|fail-message)\s(\S))")

        for line in string.splitlines():
            line_depth = depth(line)
            line = line.rstrip().lstrip()

            # skip empty lines
            if line in ["\n", "\r\n"]:
                next

            # detect banner delimiter character
            if re.match(banner_regexp, line):
                m = re.match(banner_regexp, line)
                banner_delimiter = m.group(1)
                statement = Statement(linenum, 0, [], 0, line)
                stanza.append(statement)
                del context[:]
                context.append(line.split(banner_delimiter)[0] + banner_delimiter)
                prev_line = ""
                continue

            # append banner context to end of banner
            if banner_delimiter is not None and line.startswith(banner_delimiter):
                banner_delimiter = None
                seqnum += 1
                statement = Statement(linenum, seqnum, context.copy(), 0, "")
                stanza.append(statement)
                del context[:]
                seqnum = 0
                continue

            # end-policy, end-set, end-group
            if re.match(r"^end\-", line):
                statement = Statement(linenum, seqnum, context.copy(), line_depth, line)
                stanza.append(statement)
                del context[:]
                seqnum = 0
                continue

            # if RPL we dont stack up the previous line
            if len(context) >= 1:
                if re.match(r"^route-policy", context[0]):
                    seqnum += 1

            # normal case : push and pop context on the fly
            if line_depth == 0 and banner_delimiter is None:
                del context[:]
                seqnum = 0
                current_context_depth = 0

            elif line_depth >= prev_line_depth + 1:
                context.append(prev_line)
                current_context_depth = prev_line_depth
                seqnum += 1

            elif line_depth <= prev_line_depth - 1:
                # we pop n elements depending on the diff in line level
                if current_context_depth > line_depth:
                    del context[-1:]
                    current_context_depth = line_depth
                    seqnum += 1
                # we pop only once normally
                elif line_depth == current_context_depth:
                    context.pop()
                    seqnum += 1
                # sometimes we don't pop
                else:
                    seqnum += 1
            else:
                seqnum += 1

            statement = Statement(linenum, seqnum, context.copy(), line_depth, line)

            stanza.append(statement)
            prev_line_depth = line_depth
            prev_line = line.rstrip().lstrip()
            linenum += 1

        return stanza



import argparse
import hashlib
import re
import sys

from cfgformalizer.cisco.cisco import Cisco


class RegexpAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        regexp = re.compile(value)
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = []
        if option_string == "--include":
            items.append(("include", regexp))
        if option_string == "--exclude":
            items.append(("exclude", regexp))
        setattr(namespace, self.dest, items)


class ReplacementAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        regexp = re.compile(values[0])
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = []
        items.append((regexp, values[1]))
        setattr(namespace, self.dest, items)


def get_parser():
    """Return CLI parser"""
    description = "Expand context of Cisco configuration file to make them greppable"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "--linenum",
        dest="linenum",
        help="Display original config line numbers",
        action="store_true",
    )

    parser.add_argument(
        "--seqnum",
        dest="seqnum",
        help="Display stanza sequence number",
        action="store_true",
    )

    parser.add_argument(
        "--comments", dest="comments", help="Display comments", action="store_true"
    )

    parser.add_argument(
        "--sort",
        dest="sort",
        help="Sort statements alphabetically",
        action="store_true",
    )

    parser.add_argument(
        "--delimiter",
        dest="delimiter",
        default=" ",
        help="Change context delimiter (space by default)",
        action="store",
    )

    parser.add_argument(
        "--include",
        dest="regexps",
        help="Stack regexps for filtering output",
        action=RegexpAction,
    )

    parser.add_argument(
        "--exclude",
        dest="regexps",
        help="Stack regexps for filtering output",
        action=RegexpAction,
    )

    parser.add_argument(
        "--normal",
        dest="normal",
        help="Display normal config style instead of formal",
        action="store_true",
    )

    parser.add_argument(
        "--replace",
        nargs=2,
        dest="replace",
        metavar=("REGEXP", "STRING"),
        help="Replace regexp with string",
        action=ReplacementAction,
    )

    parser.add_argument(
        "--hash",
        dest="hash",
        help="Hash stanza for easy comparison",
        action="store_true",
    )

    parser.add_argument(
        "--hash-algorithm",
        dest="hash_algo",
        help="Hash algorithm used for hashing",
        action="store",
        default="sha256",
    )

    parser.add_argument("files", nargs="+", metavar="FILE", help="Configuration file")
    return parser


def build_stanza(config, regexps=None, comments=False, sort=False):
    s = Cisco.parse(config)
    if regexps:
        for regexp_type, regexp in regexps:
            if regexp_type == "include":
                s = s.like(regexp)
            elif regexp_type == "exclude":
                s = s.unlike(regexp)
    if not comments:
        s = s.without_comments()
    if sort:
        s = s.sort()
    return s


def stringify_stanza(stanza, normal=False, linenum=False, seqnum=False, delimiter=" "):
    if normal:
        return stanza.normal(linenum=linenum, seqnum=seqnum, delimiter=delimiter)
    else:
        return stanza.formal(linenum=linenum, seqnum=seqnum, delimiter=delimiter)


def replace_pattern(string, regexp, replacement_string):
    return regexp.sub(replacement_string, string)


def main():
    parser = get_parser()
    args = parser.parse_args()
    contents = []
    if args.files:
        if args.files[0] == "-":
            contents.append(sys.stdin.read())
        else:
            for filename in args.files:
                with open(filename, "r") as f:
                    contents.append(f.read())

    for c in contents:
        s = build_stanza(
            c, regexps=args.regexps, comments=args.comments, sort=args.sort
        )

        s = stringify_stanza(
            s,
            normal=args.normal,
            linenum=args.linenum,
            seqnum=args.seqnum,
            delimiter=args.delimiter,
        )

        if args.replace:
            for regexp, replacement_string in args.replace:
                s = replace_pattern(s, regexp, replacement_string)

        if args.hash:
            h = getattr(hashlib, args.hash_algo)
            print(h(s.encode("utf-8")).hexdigest())
        else:
            print(s)


if __name__ == "__main__":
    main()



import hashlib
import re

from cfgformalizer.decorators import clone


class Stanza:
    def __init__(self, l=None):
        self.statements = l or []

    def __str__(self):
        s = [str(statement) for statement in self.statements]
        return "\n".join(s)

    def __eq__(self, other):
        return self.hash() == other.hash()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, statement):
        return statement in self.statements

    def __iter__(self):
        return iter(self.statements)

    def __len__(self):
        return len(self.statements)

    def __getitem__(self, index):
        return self.statements[index]

    def append(self, statement):
        self.statements.append(statement)

    def contexts(self):
        return [statement.context for statement in self.statements]

    @clone
    def like(self, regexp):
        self.statements = [
            statement
            for statement in self.statements
            if regexp.search(statement.formal())
        ]

    @clone
    def unlike(self, regexp):
        self.statements = [
            statement
            for statement in self.statements
            if not regexp.search(statement.formal())
        ]

    @clone
    def sort(self):
        self.statements = sorted(self.statements)

    @clone
    def without_comments(self):
        self.statements = [
            statement for statement in self.statements if not statement.is_comment()
        ]

    def normal(self, linenum=False, seqnum=False, delimiter=" "):
        s = [
            statement.normal(linenum, seqnum, delimiter)
            for statement in self.statements
        ]
        return "\n".join(s)

    def formal(self, linenum=False, seqnum=False, delimiter=" "):
        s = [
            statement.formal(linenum, seqnum, delimiter)
            for statement in self.statements
        ]
        return "\n".join(s)

    def hash(self, algorithm=None):
        a = algorithm or "sha256"
        if a not in list(hashlib.algorithms_available):
            raise Exception("algorithm {} is not supported".format(a))
        h = getattr(hashlib, a)
        return h(self.formal().encode("utf-8")).hexdigest()



def clone(func):
    def decorator(self, *args, **kwargs):
        new_self = self.__class__.__new__(self.__class__)
        new_self.statements = self.statements.copy()
        func(new_self, *args, **kwargs)
        return new_self

    return decorator



from cfgformalizer.statement import Statement


class Statement(Statement):
    def __str__(self):
        return self.normal()

    def is_comment(self):
        if len(self.string) > 0:
            return self.string[0] == "!"

    def normal(self, linenum=False, seqnum=False, delimiter=" "):
        s = []
        if linenum:
            s.append("%6s" % self.linenum)
            s.append(delimiter)
        if seqnum:
            s.append("%6s" % self.seqnum)
            s.append(delimiter)
        s.append(self.depth * delimiter + self.string)
        return "".join(s)

    def formal(self, linenum=False, seqnum=False, delimiter=" "):
        s = []
        if linenum:
            s.append("%6s" % self.linenum)
            s.append(delimiter)
        if seqnum:
            s.append("%6s" % self.seqnum)
            s.append(delimiter)
        if len(self.context) == 0:
            s.append(self.string)
            s.append(delimiter)
        else:
            s.append(delimiter.join(self.context))
            s.append(delimiter)
            s.append(self.string)
        return "".join(s)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cfgformalizer']

package_data = \
{'': ['*'], 'cfgformalizer': ['cisco/*']}

entry_points = \
{'console_scripts': ['cfgformalizer = cfgformalizer.command_line:main']}

setup_kwargs = {
    'name': 'cfgformalizer',
    'version': '0.1.0',
    'description': 'Formalize (show run formal) Cisco IOS, IOS-XR and other type of network configs for easy grepping',
    'long_description': None,
    'author': 'xv0x7c0',
    'author_email': '13061441+xv0x7c0@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class Statement:
    def __init__(self, linenum, seqnum, context, depth, string):
        self.linenum = linenum
        self.seqnum = seqnum
        self.context = context
        self.depth = depth
        self.string = string

    def __lt__(self, other):
        return self.formal() < other.formal()
